@@ -9,7 +9,6 @@
 
 [@@@warning "-27"]
 
-open CamomileLibraryDefault.Camomile
 open Lwt_react
 open LTerm_text
 open LTerm_geom
@@ -17,23 +16,17 @@ open LTerm_style
 
 let (>>=) = Lwt.(>>=)
 
-let toploop_get_directive name =
-#if OCAML_VERSION >= (4, 13, 0)
-  Toploop.get_directive name
-#else
-  try Some (Hashtbl.find Toploop.directive_table name) with Not_found -> None
-#endif
-
 module String_set = Set.Make(String)
 
-let version = "2.9.0"
+let version = "2.13.1"
 
 (* +-----------------------------------------------------------------+
    | History                                                         |
    +-----------------------------------------------------------------+ *)
+module Default_paths = UTop_private.Default_paths
 
 let history = LTerm_history.create []
-let history_file_name = ref (Some (Filename.concat LTerm_resources.home ".utop-history"))
+let history_file_name = ref (Some (Default_paths.history_file_name))
 let history_file_max_size = ref None
 let history_file_max_entries = ref None
 let stashable_session_history = UTop_history.create ()
@@ -205,10 +198,7 @@ let discard_formatters pps f =
   (* Output functions. *)
   let out_functions = {
     Format.out_string = (fun _ _ _ -> ()); out_flush = ignore;
-    out_newline = ignore; out_spaces = ignore
-#if OCAML_VERSION >= (4, 06, 0)
-      ; out_indent = ignore
-#endif
+    out_newline = ignore; out_spaces = ignore ; out_indent = ignore
   } in
   (* Replace formatter functions. *)
   List.iter (fun pp -> Format.pp_set_formatter_out_functions pp out_functions) pps;
@@ -271,14 +261,10 @@ let parse_default parse str eos_is_error =
         (* If the string is empty, do not report an error. *)
         raise Need_more
     | Lexer.Error (error, loc) ->
-#if OCAML_VERSION >= (4, 08, 0)
         (match Location.error_of_exn (Lexer.Error (error, loc)) with
         | Some (`Ok error)->
           Error ([mkloc loc], get_message Location.print_report error)
         | _-> raise Need_more)
-#else
-        Error ([mkloc loc], get_message Lexer.report_error error)
-#endif
     | Syntaxerr.Error error -> begin
       match error with
       | Syntaxerr.Unclosed (opening_loc, opening, closing_loc, closing) ->
@@ -302,10 +288,16 @@ let parse_default parse str eos_is_error =
       | Syntaxerr.Ill_formed_ast (loc, s) ->
           Error ([mkloc loc],
                  Printf.sprintf "Error: broken invariant in parsetree: %s" s)
-#if OCAML_VERSION >= (4, 03, 0)
       | Syntaxerr.Invalid_package_type (loc, s) ->
           Error ([mkloc loc],
                  Printf.sprintf "Invalid package type: %s" s)
+#if OCAML_VERSION >= (5, 0, 0)
+      | Syntaxerr.Removed_string_set loc ->
+          Error ([mkloc loc],
+            "Syntax error: strings are immutable, there is no assignment \
+             syntax for them.\n\
+             Hint: Mutable sequences of bytes are available in the Bytes module.\n\
+             Hint: Did you mean to use 'Bytes.set'?")
 #endif
     end
     | Syntaxerr.Escape_error | Parsing.Parse_error ->
@@ -338,12 +330,6 @@ let with_loc loc str = {
   Location.loc = loc;
 }
 
-#if OCAML_VERSION >= (4, 03, 0)
-let nolabel = Asttypes.Nolabel
-#else
-let nolabel = ""
-#endif
-
 (* Check that the given phrase can be evaluated without typing/compile
    errors. *)
 let check_phrase phrase =
@@ -364,20 +350,17 @@ let check_phrase phrase =
         let env = !Toploop.toplevel_env in
         (* Construct "let _ () = let module _ = struct <items> end in ()" in order to test
            the typing and compilation of [items] without evaluating them. *)
-        let unit = with_loc loc (Longident.Lident "()") in
+        let unit =
+          let (%.) a b = Longident.Ldot (a, b) in
+          with_loc loc (Lident "Stdlib" %. "Unit" %. "()")
+        in
         let top_def =
           let open Ast_helper in
           with_default_loc loc
             (fun () ->
                Str.eval
-                 (Exp.fun_ nolabel None (Pat.construct unit None)
-                   (Exp.letmodule (with_loc loc
-                        #if OCAML_VERSION >= (4, 10, 0)
-                        (Some "_")
-                        #else
-                          "_"
-                        #endif
-                        )
+                 (Exp.fun_ Nolabel None (Pat.construct unit None)
+                   (Exp.letmodule (with_loc loc (Some "_"))
                       (Mod.structure (item :: items))
                       (Exp.construct unit None))))
         in
@@ -477,7 +460,7 @@ let make_prompt ui profile count size key_sequence (recording, macro_count, macr
               txta;
               Array.make
                 (size.cols - Array.length txta - Array.length txtb)
-                (Zed_char.unsafe_of_uChar (UChar.of_int 0x2500), { none with foreground = Some (color lcyan blue); bold = Some bold });
+                (Zed_char.of_utf8 "\u{2500}", { none with foreground = Some (color lcyan blue); bold = Some bold });
               txtb;
             ]
         ) second_line
@@ -803,17 +786,9 @@ let () =
    | Backports                                                       |
    +-----------------------------------------------------------------+ *)
 
-
-let try_finally ~always work=
-#if OCAML_VERSION >= (4, 08, 0)
-    Misc.try_finally ~always work
-#else
-    Misc.try_finally work always
-#endif
-
 let use_output command =
   let fn = Filename.temp_file "ocaml" "_toploop.ml" in
-  try_finally ~always:(fun () ->
+  Misc.try_finally ~always:(fun () ->
     try Sys.remove fn with Sys_error _ -> ())
     (fun () ->
        match
@@ -828,7 +803,7 @@ let use_output command =
 
 let () =
   let name = "use_output" in
-  if toploop_get_directive name = None then
+  if UTop_compat.toploop_get_directive name = None then
     Toploop.add_directive
       name
       (Toploop.Directive_string use_output)
@@ -853,20 +828,35 @@ let () =
    | Compiler-libs re-exports                                        |
    +-----------------------------------------------------------------+ *)
 
-#if OCAML_VERSION >= (4, 08, 0)
-let get_load_path ()= Load_path.get_paths ()
-let set_load_path path= Load_path.init path
-#else
-let get_load_path ()= !Config.load_path
-let set_load_path path= Config.load_path := path
-#endif
+let get_load_path () = Load_path.get_paths ()
+let set_load_path = UTop_compat.set_load_path
 
-(* +-----------------------------------------------------------------+
-   | Deprecated                                                      |
-   +-----------------------------------------------------------------+ *)
-
-let smart_accept = ref true
-let new_prompt_hooks = LTerm_dlist.create ()
-let at_new_prompt f = ignore (LTerm_dlist.add_l f new_prompt_hooks)
-let prompt_continue = ref (S.const [| |])
-let prompt_comment = ref (S.const [| |])
+module Private = struct
+  let fix_string str =
+    let len = String.length str in
+    if len = 0 then
+      str
+    else
+      let ofs, _, _ = Zed_utf8.next_error str 0 in
+      if ofs = len then
+        str
+      else begin
+        let buf = Buffer.create (len + 128) in
+        if ofs > 0 then Buffer.add_substring buf str 0 ofs;
+        let rec loop ofs =
+          Zed_utf8.add buf (Uchar.of_char str.[ofs]);
+          let ofs1 = ofs + 1 in
+          if ofs1 < len then
+            let ofs2, _, _ = Zed_utf8.next_error str ofs1 in
+            if ofs1 < ofs2 then
+              Buffer.add_substring buf str ofs1 (ofs2 - ofs1);
+            if ofs2 < len then
+              loop ofs2
+            else
+              Buffer.contents buf
+          else
+            Buffer.contents buf
+        in
+        loop ofs
+      end
+end
