@@ -11,6 +11,7 @@
 
 open Types
 open LTerm_read_line
+open UTop_compat
 open UTop_token
 
 module String_set = Set.Make(String)
@@ -22,27 +23,6 @@ let set_of_list = List.fold_left (fun set x -> String_set.add x set) String_set.
    | Utils                                                           |
    +-----------------------------------------------------------------+ *)
 
-let get_desc x =
-#if OCAML_VERSION >= (4, 14, 0)
-  Types.get_desc x
-#else
-  x.Types.desc
-#endif
-
-let toploop_get_directive name =
-#if OCAML_VERSION >= (4, 13, 0)
-  Toploop.get_directive name
-#else
-  try Some (Hashtbl.find Toploop.directive_table name) with Not_found -> None
-#endif
-
-let toploop_all_directive_names () =
-#if OCAML_VERSION >= (4, 13, 0)
-  Toploop.all_directive_names ()
-#else
-  Hashtbl.fold (fun dir _ acc -> dir::acc) Toploop.directive_table []
-#endif
-
 (* Transform a non-empty list of strings into a long-identifier. *)
 let longident_of_list = function
   | [] ->
@@ -414,11 +394,7 @@ let visible_modules () =
               (Sys.readdir (if dir = "" then Filename.current_dir_name else dir))
           with Sys_error _ ->
             acc)
-#if OCAML_VERSION >= (4, 08, 0)
         String_set.empty @@ Load_path.get_paths ()
-#else
-        String_set.empty !Config.load_path
-#endif
     )
 
 let field_name { ld_id = id } = Ident.name id
@@ -426,11 +402,7 @@ let constructor_name { cd_id = id } = Ident.name id
 
 let add_fields_of_type decl acc =
   match decl.type_kind with
-#if OCAML_VERSION >= (4, 13, 0)
-    | Type_variant (constructors, _) ->
-#else
-    | Type_variant constructors ->
-#endif
+    | Type_variant _ ->
         acc
     | Type_record (fields, _) ->
         List.fold_left (fun acc field -> add (field_name field) acc) acc fields
@@ -454,45 +426,22 @@ let add_names_of_type decl acc =
     | Type_open ->
         acc
 
-#if OCAML_VERSION >= (4, 08, 0)
 let path_of_mty_alias = function
   | Mty_alias path -> path
   | _ -> assert false
-#elif OCAML_VERSION >= (4, 04, 0)
-let path_of_mty_alias = function
-  | Mty_alias (_, path) -> path
-  | _ -> assert false
-#else
-let path_of_mty_alias = function
-  | Mty_alias path -> path
-  | _ -> assert false
-#endif
 
 let rec names_of_module_type = function
   | Mty_signature decls ->
       List.fold_left
         (fun acc decl -> match decl with
-#if OCAML_VERSION >= (4, 08, 0)
            | Sig_value (id, _, _)
            | Sig_typext (id, _, _, _)
            | Sig_module (id, _, _, _, _)
            | Sig_modtype (id, _, _)
            | Sig_class (id, _, _, _)
            | Sig_class_type (id, _, _, _) ->
-#else
-           | Sig_value (id, _)
-           | Sig_typext (id, _, _)
-           | Sig_module (id, _, _)
-           | Sig_modtype (id, _)
-           | Sig_class (id, _, _)
-           | Sig_class_type (id, _, _) ->
-#endif
                add (Ident.name id) acc
-#if OCAML_VERSION >= (4, 08, 0)
            | Sig_type (id, decl, _, _) ->
-#else
-           | Sig_type (id, decl, _) ->
-#endif
                add_names_of_type decl (add (Ident.name id) acc))
         String_set.empty decls
   | Mty_ident path -> begin
@@ -521,11 +470,7 @@ let rec fields_of_module_type = function
            | Sig_class _
            | Sig_class_type _ ->
                acc
-#if OCAML_VERSION >= (4, 08, 0)
            | Sig_type (_, decl, _, _) ->
-#else
-           | Sig_type (_, decl, _) ->
-#endif
                add_fields_of_type decl acc)
         String_set.empty decls
   | Mty_ident path -> begin
@@ -543,24 +488,15 @@ let rec fields_of_module_type = function
   | _ ->
       String_set.empty
 
-let lookup_module id env =
-#if OCAML_VERSION >= (4, 10, 0)
-  let path, decl = Env.find_module_by_name id env in
-  (path, decl.md_type)
-#else
-  let path = Env.lookup_module id env ~load:true in
-  (path, (Env.find_module path env).md_type)
-#endif
-
 let find_module path env = (Env.find_module path env).md_type
 
 let names_of_module longident =
   try
     Longident_map.find longident !local_names_by_longident
   with Not_found ->
-    match lookup_env lookup_module longident !Toploop.toplevel_env with
-      | Some(path, module_type) ->
-          let names = names_of_module_type module_type in
+    match lookup_env Env.find_module_by_name longident !Toploop.toplevel_env with
+      | Some(path, {md_type; _}) ->
+          let names = names_of_module_type md_type in
           local_names_by_path := Path_map.add path names !local_names_by_path;
           local_names_by_longident := Longident_map.add longident names !local_names_by_longident;
           names
@@ -572,9 +508,9 @@ let fields_of_module longident =
   try
     Longident_map.find longident !local_fields_by_longident
   with Not_found ->
-    match lookup_env lookup_module longident !Toploop.toplevel_env with
-      | Some(path, module_type) ->
-          let fields = fields_of_module_type module_type in
+    match lookup_env Env.find_module_by_name longident !Toploop.toplevel_env with
+      | Some(path, {md_type; _}) ->
+          let fields = fields_of_module_type md_type in
           local_fields_by_path := Path_map.add path fields !local_fields_by_path;
           local_fields_by_longident := Longident_map.add longident fields !local_fields_by_longident;
           fields
@@ -585,21 +521,15 @@ let fields_of_module longident =
 let list_global_names () =
   let rec loop acc = function
     | Env.Env_empty -> acc
-#if OCAML_VERSION >= (4, 10, 0)
     | Env.Env_value_unbound _-> acc
     | Env.Env_module_unbound _-> acc
-#endif
     | Env.Env_value(summary, id, _) ->
         loop (add (Ident.name id) acc) summary
     | Env.Env_type(summary, id, decl) ->
         loop (add_names_of_type decl (add (Ident.name id) acc)) summary
     | Env.Env_extension(summary, id, _) ->
         loop (add (Ident.name id) acc) summary
-#if OCAML_VERSION >= (4, 08, 0)
     | Env.Env_module(summary, id, _, _) ->
-#else
-    | Env.Env_module(summary, id, _) ->
-#endif
         loop (add (Ident.name id) acc) summary
     | Env.Env_modtype(summary, id, _) ->
         loop (add (Ident.name id) acc) summary
@@ -609,28 +539,13 @@ let list_global_names () =
         loop (add (Ident.name id) acc) summary
     | Env.Env_functor_arg(summary, id) ->
         loop (add (Ident.name id) acc) summary
-#if OCAML_VERSION >= (4, 08, 0)
     | Env.Env_persistent (summary, id) ->
         loop (add (Ident.name id) acc) summary
-#endif
-#if OCAML_VERSION >= (4, 04, 0)
     | Env.Env_constraints (summary, _) ->
         loop acc summary
-#endif
-#if OCAML_VERSION >= (4, 10, 0)
     | Env.Env_copy_types summary ->
         loop acc summary
-#elif OCAML_VERSION >= (4, 06, 0)
-    | Env.Env_copy_types (summary, _) ->
-        loop acc summary
-#endif
-#if OCAML_VERSION >= (4, 08, 0)
     | Env.Env_open(summary, path) ->
-#elif OCAML_VERSION >= (4, 07, 0)
-    | Env.Env_open(summary, _, path) ->
-#else
-    | Env.Env_open(summary, path) ->
-#endif
         match try Some (Path_map.find path !local_names_by_path) with Not_found -> None with
           | Some names ->
               loop (String_set.union acc names) summary
@@ -660,21 +575,15 @@ let replace x y set =
 let list_global_fields () =
   let rec loop acc = function
     | Env.Env_empty -> acc
-#if OCAML_VERSION >= (4, 10, 0)
     | Env.Env_value_unbound _-> acc
     | Env.Env_module_unbound _-> acc
-#endif
     | Env.Env_value(summary, id, _) ->
         loop (add (Ident.name id) acc) summary
     | Env.Env_type(summary, id, decl) ->
         loop (add_fields_of_type decl (add (Ident.name id) acc)) summary
     | Env.Env_extension(summary, id, _) ->
         loop (add (Ident.name id) acc) summary
-#if OCAML_VERSION >= (4, 08, 0)
     | Env.Env_module(summary, id, _, _) ->
-#else
-    | Env.Env_module(summary, id, _) ->
-#endif
         loop (add (Ident.name id) acc) summary
     | Env.Env_functor_arg(summary, id) ->
         loop (add (Ident.name id) acc) summary
@@ -684,30 +593,13 @@ let list_global_fields () =
         loop (add (Ident.name id) acc) summary
     | Env.Env_cltype(summary, id, _) ->
         loop (add (Ident.name id) acc) summary
-#if OCAML_VERSION >= (4, 08, 0)
     | Env.Env_persistent (summary, id) ->
         loop (add (Ident.name id) acc) summary
-#endif
-#if OCAML_VERSION >= (4, 04, 0)
     | Env.Env_constraints (summary, _) ->
         loop acc summary
-#endif
-#if OCAML_VERSION >= (4, 10, 0)
     | Env.Env_copy_types summary ->
         loop acc summary
-#elif OCAML_VERSION >= (4, 06, 0)
-    | Env.Env_copy_types (summary, _) ->
-        loop acc summary
-#endif
-#if OCAML_VERSION >= (4, 07, 0)
-  #if OCAML_VERSION >= (4, 08, 0)
-      | Env.Env_open(summary, path) ->
-  #else
-      | Env.Env_open(summary, _, path) ->
-  #endif
-#else
     | Env.Env_open(summary, path) ->
-#endif
         match try Some (Path_map.find path !local_fields_by_path) with Not_found -> None with
           | Some fields ->
               loop (String_set.union acc fields) summary
@@ -789,14 +681,7 @@ let rec find_object meths type_expr =
               None
 
 let methods_of_object longident meths =
-  let lookup_value=
-#if OCAML_VERSION >= (4, 10, 0)
-    Env.find_value_by_name
-#else
-    Env.lookup_value
-#endif
-  in
-  match lookup_env lookup_value longident !Toploop.toplevel_env with
+  match lookup_env Env.find_value_by_name longident !Toploop.toplevel_env with
     | None ->
         []
     | Some (path, { val_type = type_expr }) ->
@@ -817,14 +702,6 @@ let rec labels_of_type acc type_expr =
     | Tpoly (te, _) ->
         labels_of_type acc te
     | Tarrow(label, _, te, _) ->
-#if OCAML_VERSION < (4, 03, 0)
-        if label = "" then
-          labels_of_type acc te
-        else if label.[0] = '?' then
-          labels_of_type (String_map.add (String.sub label 1 (String.length label - 1)) Optional acc) te
-        else
-          labels_of_type (String_map.add label Required acc) te
-#else
       (match label with
       | Nolabel ->
         labels_of_type acc te
@@ -832,7 +709,6 @@ let rec labels_of_type acc type_expr =
         labels_of_type (String_map.add label Optional acc) te
       | Labelled label ->
         labels_of_type (String_map.add label Required acc) te)
-#endif
     | Tconstr(path, _, _) -> begin
         match lookup_env Env.find_type path !Toploop.toplevel_env with
           | None
@@ -845,14 +721,7 @@ let rec labels_of_type acc type_expr =
         String_map.bindings acc
 
 let labels_of_function longident meths =
-  let lookup_value=
-#if OCAML_VERSION >= (4, 10, 0)
-    Env.find_value_by_name
-#else
-    Env.lookup_value
-#endif
-  in
-  match lookup_env lookup_value longident !Toploop.toplevel_env with
+  match lookup_env Env.find_value_by_name longident !Toploop.toplevel_env with
     | None ->
         []
     | Some (path, { val_type = type_expr }) ->
@@ -863,14 +732,7 @@ let labels_of_function longident meths =
               labels_of_type String_map.empty type_expr
 
 let labels_of_newclass longident =
-  let lookup_class=
-#if OCAML_VERSION >= (4, 10, 0)
-    Env.find_class_by_name
-#else
-    Env.lookup_class
-#endif
-  in
-  match lookup_env lookup_class longident !Toploop.toplevel_env with
+  match lookup_env Env.find_class_by_name longident !Toploop.toplevel_env with
     | None ->
         []
     | Some (path, { cty_new = None }) ->
@@ -922,14 +784,6 @@ and find_context_in_quotation = function
    | Completion                                                      |
    +-----------------------------------------------------------------+ *)
 
-#if OCAML_VERSION < (4, 11, 0)
-let longident_parse= Longident.parse
-#else
-let longident_parse str=
-  let lexbuf= Lexing.from_string str in
-  Parse.longident lexbuf
-#endif
-
 let complete ~phrase_terminator ~input =
   let true_name, false_name = ("true", "false") in
   let tokens = UTop_lexer.lex_string input in
@@ -962,7 +816,7 @@ let complete ~phrase_terminator ~input =
 
     | [(Symbol "#", _); (Lident "typeof", _); (String (tlen, false), loc)] ->
       let prefix = String.sub input (loc.ofs1 + tlen) (String.length input - loc.ofs1 - tlen) in
-      begin match longident_parse prefix with
+      begin match Parse.longident (Lexing.from_string prefix) with
       | Longident.Ldot (lident, last_prefix) ->
         let set = names_of_module lident in
         let compls = lookup last_prefix (String_set.elements set) in
@@ -985,11 +839,7 @@ let complete ~phrase_terminator ~input =
               (fun acc d -> add_files filter acc (Filename.concat d dir))
               String_map.empty
               (Filename.current_dir_name ::
-#if OCAML_VERSION >= (4, 08, 0)
                 (Load_path.get_paths ())
-#else
-                !Config.load_path
-#endif
               )
 
           else
@@ -1049,11 +899,7 @@ let complete ~phrase_terminator ~input =
               (fun acc d -> add_files filter acc (Filename.concat d dir))
               String_map.empty
               (Filename.current_dir_name ::
-#if OCAML_VERSION >= (4, 08, 0)
                 (Load_path.get_paths ())
-#else
-                !Config.load_path
-#endif
               )
           else
             add_files filter String_map.empty (Filename.dirname file)
